﻿/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// .NET Source Code
// Copyright © 2012 MobilEyze, LLC
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// Author	Date		Comment
// =========================================================================================================
// NLW      12 Feb 2012 Created 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace MobilEyze.Tools.DataLogger.ToolSuite.Model
{
    /// <summary>
    /// This class is responsible for reading and writing MobilEyze Data Logger Configuration Files.
    /// </summary>
    public class MEConfigurationFile
    {
        // Maximum ID's
        public static readonly uint MAX_11BIT_ID = 0x1EF;
        public static readonly uint MAX_29BIT_ID = 0x1FBFFFFF;

        #region Supporting Enums

        public enum E_DATALOGGER_BUS_SPEED
        {
            BAUD_100K,
            BAUD_125K,
            BAUD_250K,
            BAUD_500K,
            BAUD_1M,
        }
        public enum E_DATALOGGER_DEBUG
        {
            None,
            Serial,
            CAN,
        }
        public enum E_CAN_MSG_TYPE
        {
            Standard,
            Extended
        }
        public enum E_LogType
        {
            Standard,
            Circular,
        }
        #endregion

        #region Supporting classes

        /// <summary>
        /// A class representing a valid CAN filter that can be written to the config file.
        /// </summary>
        public class CANFilter
        {
            public uint StartingMessageID { get; set; }
            public uint EndingMessageID { get; set; }
            public E_CAN_MSG_TYPE MessageType { get; set; }
        }
        #endregion

        #region Constructor

        public MEConfigurationFile()
        {
            this.Filters = new CANFilter[MAX_FILTERS];
            this.UseErrorPassiveDetection = false;
            this.CANId = 0x0FE;
        }

        private MEConfigurationFile(string fileName) : this()
        {
            this.FileName = fileName;
        }
        #endregion

        #region Public Properties

        public const uint MAX_FILTERS = 10;
        public string FileName { get; private set; }
        public E_DATALOGGER_BUS_SPEED BusSpeed { get; private set; }
        public CANFilter[] Filters { get; internal set; }
        public bool UseErrorPassiveDetection { get; set; }
        public uint CANId { get; set; }
        public bool OutputDebugToCAN { get; set; }
        public E_LogType LogType { get; set; }

        #endregion

        #region Clear Filters
        /// <summary>
        /// Clear the existing filters.
        /// </summary>
        public void ClearFilters()
        {
            this.Filters = new CANFilter[MAX_FILTERS];
        }

        #endregion

        #region Add Filter
        /// <summary>
        /// Add a new filter
        /// </summary>
        /// <param name="startingID"></param>
        /// <param name="endingID"></param>
        /// <param name="CAN_msg_type"></param>
        /// <returns></returns>
        public bool AddFilter(uint startingID, uint endingID, E_CAN_MSG_TYPE CAN_msg_type)
        {
            if (CAN_msg_type == E_CAN_MSG_TYPE.Standard)
            {
                if (startingID > endingID)
                {
                    throw new ArgumentException("starting id can't be greater than ending id.");
                }

                if (startingID > MAX_11BIT_ID || endingID > MAX_11BIT_ID)
                {
                    throw new ArgumentException(string.Format("11bit id's can't be greater than {0}", MAX_11BIT_ID));
                }
            }
            else
            {
                if (startingID > endingID)
                {
                    throw new ArgumentException("starting id can't be greater than ending id.");
                }

                if (startingID > MAX_29BIT_ID || endingID > MAX_29BIT_ID)
                {
                    throw new ArgumentException(string.Format("11bit id's can't be greater than {0}", MAX_29BIT_ID));
                }
            }
            // Find an open slot
            for (int i = 0; i < MAX_FILTERS; i++)
            {
                if (this.Filters[i] == null)
                {
                    this.Filters[i] = new CANFilter();
                    this.Filters[i].StartingMessageID = startingID;
                    this.Filters[i].EndingMessageID = endingID;
                    this.Filters[i].MessageType = CAN_msg_type;
                    return true;
                }
            }

            return false;
        }

        #endregion

        #region Set Bus Speed
        /// <summary>
        /// Set the desired bus speed
        /// </summary>
        /// <param name="busSpeed"></param>
        public void SetBusSpeed(string busSpeed)
        {
            this.BusSpeed = convertBusSpeed(busSpeed);
        }

        public static string GetBusSpeedString(E_DATALOGGER_BUS_SPEED busSpeed)
        {
            return convertBusSpeed(busSpeed);
        }
        #endregion

        #region Set Output Debug To CAN Option
        /// <summary>
        /// Determine if the modules debug messages should be output via CAN - Use at your own risk!
        /// </summary>
        /// <param name="ouputDebugToCAN"></param>
        public void SetOutputDebugToCAN(bool ouputDebugToCAN)
        {
            this.OutputDebugToCAN = ouputDebugToCAN;
        }
        #endregion

        #region Set Log Type

        internal void SetLogType(string logType)
        {
            this.LogType = (E_LogType)Enum.Parse(typeof(E_LogType), logType);
        }
        #endregion

        #region Create

        /// <summary>
        /// Create a new configuration file.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static MEConfigurationFile Create(string fileName)
        {
            MEConfigurationFile file = new MEConfigurationFile(fileName);
            file.BusSpeed = E_DATALOGGER_BUS_SPEED.BAUD_250K;
            
            file.AddFilter(0x0, MAX_11BIT_ID, E_CAN_MSG_TYPE.Standard);
            file.AddFilter(0x0, MAX_29BIT_ID, E_CAN_MSG_TYPE.Extended);
            file.LogType = E_LogType.Standard;
            file.OutputDebugToCAN = false;
            return file;
        }

        #endregion

        #region Load
        /// <summary>
        /// Load an existing configuration file
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static MEConfigurationFile Load(string fileName)
        {
            MEConfigurationFile meConfigFile = new MEConfigurationFile();
            meConfigFile.FileName = fileName;

            // Open the file
            if (File.Exists(fileName))
            {
                FileStream fs = File.Open(fileName, FileMode.Open);
                StreamReader reader = new StreamReader(fs);

                // Read the contents
                // Contents can be:
                //  BusSpeed= [ 100K | 125K | 250K | 500K | 1M ]
                //  Filter==  [ start_msg_id,end_msg_id, ext_flag ]
                //  DebugOutput= [ S=Serial | C=CAN ]
                while (!reader.EndOfStream)
                {
                    string record = reader.ReadLine();
                    string[] array = record.Split('=');

                    if (array.Length == 2)
                    {
                        switch (array[0].ToUpper())
                        {
                            case "BUSSPEED":
                                meConfigFile.BusSpeed = convertBusSpeed(array[1]);
                                break;

                            case "FILTER":
                                string[] filterItems = array[1].Split(',');
                                if (filterItems.Length == 3)
                                {
                                    meConfigFile.AddFilter(
                                        Convert.ToUInt32(filterItems[0],16), 
                                        Convert.ToUInt32(filterItems[1],16), 
                                        convertCANMsgType(filterItems[2]));
                                }
                                break;

                            case "OUTPUTDEBUGTOCAN":
                                meConfigFile.OutputDebugToCAN = array[1].ToUpper() == "T" ? true : false;
                                break;

                            case "USEERRORPASSIVEDETECTION" :
                                meConfigFile.UseErrorPassiveDetection = array[1].ToUpper() == "T" ? true : false;
                                break;

                            case "CANID":
                                meConfigFile.CANId = Convert.ToUInt32(array[1], 16);
                                break;

                            case "LOGTYPE":
                                meConfigFile.LogType = array[1].ToUpper() == "C" ? E_LogType.Circular : E_LogType.Standard;
                                break;
                        }
                    }

                }

                reader.Close();
            
                // Return the results
                return meConfigFile;
            }

            return null;

        }

        #endregion

        #region Save
        /// <summary>
        /// Save the configuration file.
        /// </summary>
        /// <param name="meConfigFile"></param>
        public static void Save(MEConfigurationFile meConfigFile)
        {
            string fileName = meConfigFile.FileName;
            FileStream fs = File.Open(fileName, FileMode.Create);
            StreamWriter writer = new StreamWriter(fs);

            // Write the values

            // Bus Speed
            writer.WriteLine(string.Format("BusSpeed={0}", convertBusSpeed(meConfigFile.BusSpeed)));

            // Filters
            for(int i=0; i< MAX_FILTERS; i++)
            {
                CANFilter filter = meConfigFile.Filters[i];
                if (filter  != null)
                {
                    writer.WriteLine(string.Format("Filter=0x{0:x},0x{1:x},{2}", filter.StartingMessageID, filter.EndingMessageID, convertCANMsgType(filter.MessageType)));
                }
            }
            // Log Type
            writer.WriteLine(string.Format("LogType={0}", meConfigFile.LogType == E_LogType.Circular ? "C" : "S"));

            // Debug Output
            writer.WriteLine(string.Format("OutputDebugToCAN={0}", meConfigFile.OutputDebugToCAN ? "T" : "F"));

            // Error Passive Detection requires a CAN Id 
            if (meConfigFile.UseErrorPassiveDetection && meConfigFile.CANId > 0)
            {
                writer.WriteLine("UseErrorPassiveDetection=T");
                writer.WriteLine(string.Format("CANId=0x{0:x}", meConfigFile.CANId));
            }
            writer.Close();
        }

        #endregion

        #region Helpers

        public static List<string> GetSupportedBusSpeeds()
        {
            List<string> list = new List<string>();

            list.Add(convertBusSpeed(E_DATALOGGER_BUS_SPEED.BAUD_100K));
            list.Add(convertBusSpeed(E_DATALOGGER_BUS_SPEED.BAUD_125K));
            list.Add(convertBusSpeed(E_DATALOGGER_BUS_SPEED.BAUD_250K));
            list.Add(convertBusSpeed(E_DATALOGGER_BUS_SPEED.BAUD_500K));
            list.Add(convertBusSpeed(E_DATALOGGER_BUS_SPEED.BAUD_1M));
            return list;
        }

        private static  E_DATALOGGER_BUS_SPEED convertBusSpeed(string busSpeed)
        {
            switch (busSpeed)
            {
                case "100K":
                    return E_DATALOGGER_BUS_SPEED.BAUD_100K;

                case "125K":
                    return E_DATALOGGER_BUS_SPEED.BAUD_125K;

                case "250K":
                    return E_DATALOGGER_BUS_SPEED.BAUD_250K;

                case "500K":
                    return E_DATALOGGER_BUS_SPEED.BAUD_500K;

                case "1M":
                    return E_DATALOGGER_BUS_SPEED.BAUD_1M;
            }

            throw new ArgumentException(string.Format("Unable to convert {0} to an E_DATALOGGER_BUS_SPEED value.", busSpeed));
        }

        private static string convertBusSpeed(E_DATALOGGER_BUS_SPEED busSpeed)
        {
            switch (busSpeed)
            {
                case E_DATALOGGER_BUS_SPEED.BAUD_100K:
                    return "100K";

                case E_DATALOGGER_BUS_SPEED.BAUD_125K:
                    return "125K";

                case E_DATALOGGER_BUS_SPEED.BAUD_250K:
                    return "250K";

                case E_DATALOGGER_BUS_SPEED.BAUD_500K:
                    return "500K";

                case E_DATALOGGER_BUS_SPEED.BAUD_1M:
                    return "1M";
            }

            throw new ArgumentException(string.Format("Unable to convert {0} to a string value.", busSpeed));
        }

        private static E_CAN_MSG_TYPE convertCANMsgType(string msgType)
        {
            switch (msgType)
            {
                case "E":
                    return E_CAN_MSG_TYPE.Extended;

                case "S":
                    return E_CAN_MSG_TYPE.Standard;
            }
            throw new ArgumentException(string.Format("Unable to convert {0} to a valid E_CAN_MSG_TYPE value.", msgType));
        }

        private static string convertCANMsgType(E_CAN_MSG_TYPE msgType)
        {
            switch (msgType)
            {
                case E_CAN_MSG_TYPE.Extended:
                    return "E";

                case E_CAN_MSG_TYPE.Standard:
                    return "S";

            }

            throw new ArgumentException(string.Format("Unable to convert {0} to a valid string value.", msgType));
        }

        public static List<string> GetSupportedLogTypes()
        {
            return new List<string>() 
                {
                    "Standard",
                    "Circular",
                };
        }


        public static string GetLogTypeString(E_LogType e_LogType)
        {
            return Enum.GetName(typeof(E_LogType), e_LogType);
        }

        #endregion

    }
}
