﻿/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// .NET Source Code
// Copyright © 2012 MobilEyze, LLC
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// Author	Date		Comment
// =========================================================================================================
// NLW      12 Feb 2012 Created 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.Windows.Data;
using System.Diagnostics;
using System.ComponentModel;
using System.Windows.Input;
using MobilEyze.Tools.DataLogger.ToolSuite.Utilities;
using System.Windows.Shapes;

namespace MobilEyze.Tools.DataLogger.ToolSuite.View_Models
{
    /// <summary>
    /// The main view model for the application.
    /// </summary>
    public class MainVM : ViewModelBase
    {

        #region private variables
        #endregion

        #region Constructor
        /// <summary>
        /// The main view model for the application.  
        /// </summary>
        public MainVM()
        {
            this.DisplayName = "MainVM";
            //this.AppTrace = AppTraceVM.GetInstance();
            this.Workspaces = new ObservableCollection<WorkspaceVM>();

            //this.AppTrace.AddMessage(this.DisplayName, "Initialization complete.", E_AppTraceSeverity.Information);
            this.StatusBarMessage = "Ready";
        }

        #endregion

        #region App Trace
        
        //public AppTraceVM AppTrace { get; private set; }

        #endregion

        #region Status Bar Message

        private string _StatusBarMessage = string.Empty;
        public string StatusBarMessage
        {
            get { return _StatusBarMessage; }
            set
            {
                _StatusBarMessage = value;
                FirePropertyChangedEvent("StatusBarMessage");
            }
        }
        #endregion

        #region Status Bar Progress Bar
        
        public bool ShowStatusBarProgressBar { get; private set; }

        #endregion

        #region Create new configuration logic
        /// <summary>
        /// The command that creates a new configuration file
        /// </summary>
        public ICommand CreateNewConfigurationFile { get { return new RelayCommand(parm => this.OnCreateNewConfigurationFile()); } }
        private void OnCreateNewConfigurationFile()
        {
            // Prompt for the file
            string path = this.AppServices.SaveFileDialog(string.Empty, "MobileEyze Config Files (.mec)|*.mec", ".mec");
            if (!string.IsNullOrEmpty(path))
            {
                ConfigurationFileVM vm = new ConfigurationFileVM();
                vm.CreateNewConfigFile(path);

                // We want to be notified when the VM closes so that we can remove it from our workspaces
                vm.RequestClose += new EventHandler(vm_RequestClose);
                this.Workspaces.Add(vm);

                SetActiveWorkspace(vm);
                this.StatusBarMessage = "New configuration file created.";
#if DEBUG
            //this.AppTrace.AddMessage(this.DisplayName, string.Format("Add new workspace named {0}.", vm.DisplayName), E_AppTraceSeverity.Information);
#endif
            }
        }


        #endregion

        #region Open a config file
        /// <summary>
        /// The command for opening existing configuration files.
        /// </summary>
        public ICommand OpenConfigurationFile { get { return new RelayCommand(parm => this.OnOpenConfigurationFile()); } }
        private void OnOpenConfigurationFile()
        {
            // Prompt for the file
            string path = this.AppServices.OpenFileDialog(string.Empty, "MobileEyze Config Files (.mec)|*.mec", ".mec");
            if (!string.IsNullOrEmpty(path))
            {
                ConfigurationFileVM vm = new ConfigurationFileVM();
                vm.LoadConfigurationFile(path);

                // We want to be notified when the VM closes so that we can remove it from our workspaces
                vm.RequestClose += new EventHandler(vm_RequestClose);
                this.Workspaces.Add(vm);

                SetActiveWorkspace(vm);
                this.StatusBarMessage = "The configuration file is open.";
#if DEBUG
                //this.AppTrace.AddMessage(this.DisplayName, string.Format("Open new workspace named {0}.", vm.DisplayName), E_AppTraceSeverity.Information);
#endif
            }

        }
        #endregion

        #region Set Module Date and Time
        /// <summary>
        /// The command for setting the modules date and time - can only have one instance
        /// </summary>
        public ICommand SetModuleDateTime { get { return new RelayCommand(parm => this.OnSetModuleDateTime()); } }
        private void OnSetModuleDateTime()
        {
            // Only one date and time VM allowed.
            foreach (ViewModelBase vmb in this.Workspaces)
            {
                if (vmb.GetType() == typeof(SetModuleDateTimeVM))
                {
                    return;
                }
            }
            SetModuleDateTimeVM vm = new SetModuleDateTimeVM();

            // We want to be notified when the VM closes so that we can remove it from our workspaces
            vm.RequestClose += new EventHandler(vm_RequestClose);
            this.Workspaces.Add(vm);
            SetActiveWorkspace(vm);
            this.StatusBarMessage = "";
#if DEBUG
            //this.AppTrace.AddMessage(this.DisplayName, string.Format("Open new workspace named {0}.", vm.DisplayName), E_AppTraceSeverity.Information);
#endif

        }
        #endregion

        #region Handler for Workspace Close Events
        /// <summary>
        /// Event handler for when the workspaces request to be closed.
        /// </summary>
        /// <param name="sender">Must be of type WorksapceVM</param>
        /// <param name="e">is always null</param>
        void vm_RequestClose(object sender, EventArgs e)
        {
            WorkspaceVM vm = sender as WorkspaceVM;
            if (vm != null)
            {
#if DEBUG
                //this.AppTrace.AddMessage(this.DisplayName, string.Format("{0} workspace has been removed.", vm.DisplayName), E_AppTraceSeverity.Information);
#endif

                // Unregister the event
                vm.RequestClose -= new EventHandler(vm_RequestClose);

                // Remove the vm from the workspace
                this.Workspaces.Remove(vm);

                // Dispose
                vm.Dispose();

                this.StatusBarMessage = "Workspace has been removed.";
            }
        }

        #endregion

        #region Workspaces

        public ObservableCollection<WorkspaceVM> Workspaces { get; private set; }
        
        /// <summary>
        /// Make a given worksapce the active view in the UI
        /// </summary>
        /// <param name="workspace"></param>
        void SetActiveWorkspace(WorkspaceVM workspace)
        {
            Debug.Assert(this.Workspaces.Contains(workspace));

            ICollectionView collectionView = CollectionViewSource.GetDefaultView(this.Workspaces);
            if (collectionView != null)
                collectionView.MoveCurrentTo(workspace);
        }
        #endregion

    }
}
