﻿/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// .NET Source Code
// Copyright © 2012 MobilEyze, LLC
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// Author	Date		Comment
// =========================================================================================================
// NLW      12 Feb 2012 Created 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO.Ports;
using System.Threading;
using System.Diagnostics;

namespace MobilEyze.Tools.DataLogger.ToolSuite.View_Models.SetDateTime
{
    /// <summary>
    /// Handles setting the modules date and time through a Serial Interface
    /// </summary>
    public class SDTSerial : SDTBase
    {
        #region Private variables

        private SerialPort _SerialPort = null;
        private Thread _ReceiveDateTimeThread = null;
        private volatile bool _AbortThreads = false;

        #endregion

        #region Constructor

        /// <summary>
        /// Constructor for the Set Date Time via Serial interface.
        /// </summary>
        public SDTSerial()
        {
            // Build the list of available ports
            this.AvailablePortNames = new List<string>(SerialPort.GetPortNames().OrderBy(port => Convert.ToInt32(port.Replace("COM", string.Empty))));
        }

        #endregion

        #region Open
        /// <summary>
        /// Open the object.
        /// </summary>
        /// <param name="portName"></param>
        public override void Open(string portName)
        {
            // Make sure we close any previous instance
            this.Close();

            _SerialPort = new SerialPort(portName);
            _SerialPort.BaudRate = 115200;
            _SerialPort.Parity = Parity.None;
            _SerialPort.DataBits = 8;
            _SerialPort.StopBits = StopBits.One;
            _SerialPort.Open();

            // Get the receive thread up and running
            _AbortThreads = false;
            _ReceiveDateTimeThread = new Thread(new ParameterizedThreadStart(ReceiveDateTimeRequestThread));
            _ReceiveDateTimeThread.IsBackground = true;
            _ReceiveDateTimeThread.Name = "ReceiveDateTimeThread";
            _ReceiveDateTimeThread.Start(_SerialPort);
        }

        #endregion

        #region Close
        /// <summary>
        /// Close the object.
        /// </summary>
        public override void Close()
        {
            // Terminate thread
            if (_ReceiveDateTimeThread != null)
            {
                // The use of the _AbortThreads method eliminates Exception output and ensures for 
                // a better shutdown mechanism.
                _AbortThreads = true;
                _ReceiveDateTimeThread.Join();
                _ReceiveDateTimeThread = null;

            }

            // Free the serial port
            if (_SerialPort != null)
            {
                _SerialPort.Close();
                _SerialPort = null;
            }
        }

        #endregion

        #region Set Module Date and Time
        /// <summary>
        /// Set the module's date and time using a Serial connection.
        /// </summary>
        public override void SetModuleDateTimeCurrent()
        {
            // Get the current time and convert it to epoc time
            TimeSpan t = (DateTime.Now - new DateTime(1970, 1, 1));
            int timestamp = (int)t.TotalSeconds;
            _SerialPort.WriteLine(string.Format("SetDateTime={0}", timestamp));

        }

        /// <summary>
        /// Set the module's date and time using a Serial connection.
        /// </summary>
        public override void SetModuleDateTimeGMT()
        {
            // Get the current time and convert it to epoc time
            TimeSpan t = (DateTime.UtcNow - new DateTime(1970, 1, 1));
            int timestamp = (int)t.TotalSeconds;
            _SerialPort.WriteLine(string.Format("SetDateTime={0}", timestamp));

        }

        #endregion

        #region Thread

        private void ReceiveDateTimeRequestThread(object parm)
        {
            SerialPort sp = parm as SerialPort;
            if (sp == null)
            {
                return;
            }

            StringBuilder sb = new StringBuilder();
            char[] buf = new char[1];
            bool bReadData = false;
            while (!_AbortThreads)
            {
                DateTime start = DateTime.Now;
                DateTime end = DateTime.Now;
                // Send the request
                lock (sp)
                {
                    _SerialPort.Write("SendDateTime\n");
                }

                // Wait for the response
                bool bResponse = false;
                while (!bResponse)
                {
                    lock (sp)
                    {
                        try
                        {
                            bReadData = false;
                            sp.Read(buf, 0, 1);
                            bReadData = true;
                        }
                        catch { }
                    }

                    if (bReadData)
                    {
                        switch (buf[0])
                        {
                            case '\n':
                                if (sb.ToString().Contains("CurrentDateTime"))
                                {
                                    string[] arr = sb.ToString().Split('=');
                                    if (arr.Length == 2)
                                    {
                                        // Update the current date and time
                                        string sDateVal = arr[1];
                                        DateTime dtVal;
                                        DateTime.TryParse(sDateVal, out dtVal);
                                        FireModuleDateTimeChangedEvent(dtVal);
                                        Debug.WriteLine(string.Format("Current Module Time Is: {0} start={1}, end={2}, elapsed={3}", dtVal.ToLongDateString(), start, end, end.Subtract(start)));
                                        end = DateTime.Now;

                                    }
                                }
                                else
                                {
                                    Debug.WriteLine("Got partial command " + sb.ToString());
                                }
                                sb.Clear();
                                bResponse = true;
                                break;

                            default:
                                sb.Append(buf);
                                break;
                        }
                    }
                }

                Thread.Sleep(1000);
            }

#if DEBUG
            Debug.WriteLine(string.Format("SDTSerial:ReceiveDateTimeRequestThread - Thread is Aborting, _AbortThread={0}.", _AbortThreads));
#endif

        }

        #endregion
    }
}
