﻿/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// .NET Source Code
// Copyright © 2012 MobilEyze, LLC
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// Author	Date		Comment
// =========================================================================================================
// NLW      12 Feb 2012 Created 
/////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.IO.Ports;
using System.Threading;
using System.Diagnostics;
using System.Windows.Threading;
using System.Windows.Input;
using MobilEyze.Tools.DataLogger.ToolSuite.Utilities;
using MobilEyze.Tools.DataLogger.ToolSuite.View_Models.SetDateTime;

namespace MobilEyze.Tools.DataLogger.ToolSuite.View_Models
{
    /// <summary>
    /// Set the date and time VM.
    /// </summary>
    public class SetModuleDateTimeVM : WorkspaceVM
    {
        #region private variables

        Dispatcher _Dispather = null;

        #endregion

        #region Constructor
        
        public SetModuleDateTimeVM()
        {
            _Dispather = App.Current.Dispatcher;
            this.DisplayName = "Set Module Date/Time";
            this.CurrentDateTime = "Unknown";
            this.AvailablePorts = new ObservableCollection<string>();
            this.StatusMessage = "Choose the communications method.";
        }
        #endregion

        #region Public Properties
        /// <summary>
        /// The status message to be displayed to the user.
        /// </summary>
        public string StatusMessage { get; set; }

        /// <summary>
        /// The list of available communications methods.  Currently limited to CAN.
        /// </summary>
        public List<string> AvailableCommunicationMethods 
        {
            get 
            {
                List<string> lst =  new List<string> ();
                lst.Add("CAN");

                return lst;
            }
        }

        private string _DesiredCommunicationMethod = string.Empty;
        /// <summary>
        /// The desired communication method.
        /// </summary>
        public string DesiredCommunicationMethod
        {
            get { return _DesiredCommunicationMethod; }
            set
            {
                if (_DesiredCommunicationMethod != value)
                {
                    _DesiredCommunicationMethod = value;

                    this.AvailablePorts.Clear();

                    this.StatusMessage = "Choose the port.";
                    FirePropertyChangedEvent("StatusMessage");

                    // Get the availale ports
                    if (_SDT != null)
                    {
                        _SDT.Close();
                        _SDT = null;
                    }
                    switch (_DesiredCommunicationMethod)
                    {
                        case "CAN":
                            _SDT = new SDTCAN();
                            break;

                    }

                    foreach(string portName in _SDT.AvailablePortNames)
                    {
                        this.AvailablePorts.Add(portName);
                    }
                }
            }

        }

        /// <summary>
        /// The modules current date and time.  
        /// </summary>
        public string CurrentDateTime { get; set; }

        /// <summary>
        /// A list of available ports related to the communications method choosen.
        /// </summary>
        public ObservableCollection<string> AvailablePorts { get; set; }

        private string _SelectedPort = string.Empty;
        /// <summary>
        /// The selected communications port.
        /// </summary>
        public string SelectedPort
        {
            get { return _SelectedPort; }
            set
            {
                _SelectedPort = value;
                
                if (string.IsNullOrEmpty(_SelectedPort))
                {
                    return;
                }

                this.StatusMessage = "Opening the port...";
                FirePropertyChangedEvent("StatusMessage");

                // Now lets get the next one up and running
                try
                {
                    _SDT.ModuleDateTimeChanged += new EventHandler<ModuleDateTimeChangedEventArgs>(_SDT_ModuleDateTimeChanged);
                    _SDT.Open(_SelectedPort);

                    this.StatusMessage = "Port Open, waiting for date/time....";
                    FirePropertyChangedEvent("StatusMessage");

                }
                catch
                {
                    this.StatusMessage = "Error - couldn't open the port.";
                    FirePropertyChangedEvent("StatusMessage");
                }
            }
        }

        #endregion

        #region Set Date Time Interface

        /// <summary>
        /// THe object responsible for communicating with the module based on the selected communications methods.
        /// </summary>
        private ISDT _SDT { get; set; }

        /// <summary>
        /// Handler for receiving date/time updates and then updating the user interface with the new value.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _SDT_ModuleDateTimeChanged(object sender, ModuleDateTimeChangedEventArgs e)
        {
            // Get back onto the main thread
            _Dispather.BeginInvoke(new UpdateDateTimeFromModileDelegate(OnUPdateDateTimeFromModule), new object[] { e.CurrentDateTime });

        }

        #endregion

        #region Update Date and Time Value from a non ui thread

        /// <summary>
        /// Delegate used to dispatch commands back to the UI thread.
        /// </summary>
        /// <param name="dateTime"></param>
        delegate void UpdateDateTimeFromModileDelegate(DateTime dateTime);

        /// <summary>
        /// Handler for the task of updating the date and time.  This is called from _SDT_ModuleDateTimeChanged through the _Dispatcher.BeginInvoke.
        /// </summary>
        /// <param name="dateTime"></param>
        private void OnUPdateDateTimeFromModule(DateTime dateTime)
        {
            CurrentDateTime = dateTime.ToString();
            FirePropertyChangedEvent("CurrentDateTime");
        }
        #endregion

        #region Set Module Date and Time Command
        /// <summary>
        /// Command for setting the date and time based on the system date/time
        /// </summary>
        public ICommand SetModuleDateTimeCurrent { get { return new RelayCommand(parm => this.OnSetModuleDateTimeCurrent(), parm => IsOkToSetModuleDateTime()); } }

        /// <summary>
        /// Command for setting the date and time based on GMT time.
        /// </summary>
        public ICommand SetModuleDateTimeGMT { get { return new RelayCommand(parm => this.OnSetModuleDateTimeGMT(), parm => IsOkToSetModuleDateTime()); } }
        
        /// <summary>
        /// This is the method that is actually called when the command button is pressed.
        /// </summary>
        private void OnSetModuleDateTimeCurrent()
        {
            // Send the command
            lock (_SDT)
            {
                _SDT.SetModuleDateTimeCurrent();
                this.StatusMessage = "Command sent.";
                FirePropertyChangedEvent("StatusMessage");
            }
        }

        /// <summary>
        /// This is the method that is actually called when the command button is pressed.
        /// </summary>
        private void OnSetModuleDateTimeGMT()
        {
            // Send the command
            lock (_SDT)
            {
                _SDT.SetModuleDateTimeGMT();
                this.StatusMessage = "Command sent.";
                FirePropertyChangedEvent("StatusMessage");
            }
        }

        /// <summary>
        /// Determine if the set buttons should be enabled.
        /// </summary>
        /// <returns></returns>
        private bool IsOkToSetModuleDateTime()
        {
            return (_SDT != null && !string.IsNullOrEmpty(this.SelectedPort));
        }
        #endregion

        #region Workspace VM Implementation

        /// <summary>
        /// Nothing to do.
        /// </summary>
        /// <returns></returns>
        protected override bool OnSaving()
        {
            // Nothing to do
            return true;
        }

        /// <summary>
        /// Shut down the VM and free it's resources.
        /// </summary>
        protected override void OnClosing()
        {
            FreeTheSDT();
        }

        #region Overrides
        /// <summary>
        /// Allow the object to be disposed if needed.
        /// </summary>
        public override void Dispose()
        {
            this.OnClosing();
        }

        #endregion
        #endregion

        #region Helpers
        /// <summary>
        /// Free up the SDT (Set Date Time object).
        /// </summary>
        private void FreeTheSDT()
        {
            if (_SDT != null)
            {
                _SDT.ModuleDateTimeChanged -= new EventHandler<ModuleDateTimeChangedEventArgs>(_SDT_ModuleDateTimeChanged);
                _SDT.Close();
                _SDT = null;
            }
        }
        #endregion


    }

}
